################################################################
## Implementation of the Stepwise selection method proposed from Huang and Yang.
## INPUT:  'dat'  - data.frame with lags and response at first place.
##         'd'  - number of candidate covariates.
##         'Smax' - maximum number of covariates in one model.
##         'basis' - either 'ps' or 'cr'
##         'fPath' - print the forward procedure.
## OUTPUT: A subset of all covariates with minimized BIC.
################################################################

stepwise <- function(dat, Smax=5, basis="ps", fPath = F) {
  if(all(basis!=c("cr","ps")))
    stop("'basis' must be 'cr' or 'ps'.")
  if(Smax > ncol(dat)-1)
    stop("'Smax' should not be larger than 'max.lag'.")
  
  endF <- forward(dat, Smax = Smax, basis, fPath)
  fset <- endF[[Smax]]$set
  dat2 <- dat[,c(names(dat)[1],fset), drop = F]
  endB <- backward(dat2, basis, fPath)
  final<- c(endF,endB)
  BICs <- sapply(final, function(zl) zl$BIC)
  index <- which.min(BICs)
  result <- final[[index]]
  class(result) <- "stepwise_HaY"
  result
}

################################################################
## NOTE:  'forward' doesn't compute intercept-model. 
## Reason: this is a job of 'backward'.
################################################################
forward <- function(dat, Smax, basis, fPath = T){
  n   <- NROW(dat)
  resp<- names(dat)[1]
  set <- names(dat)[-1]	
  forwardset <- NULL
  endF <- vector("list", Smax)
  count <- 1
  while(length(forwardset) < Smax){
    models <- vector("list", length(set))
    k <- 1
    for(z in set)  { # z = "Series1.L2"
      predictors <- c(forwardset, z)
      work <- buildModel(resp, predictors, basis, n)
      models[[k]] <- calc.MSE_BIC(work, dat = dat)
      k <- k + 1
    }
    
    MSEs <- sapply(models, function(zl) zl$MSE)
    ind  <- which.min(MSEs)
    forwardset    <- models[[ind]]$set
    endF[[count]] <- models[[ind]]
    count <- count + 1
    set <- set[!(set %in% forwardset)]
    if(fPath)
      cat(forwardset, "\n") 
  } 
  endF
}

##############################################################
# NOTE:  'backward' doesn't compute full-model and 'single-term model'. 
# Reason: this is a job of 'forward'.
##############################################################
backward <- function(dat, basis, fPath = F){
  n <- nrow(dat)
  resp <- names(dat)[1]
  set  <- names(dat)[-1]
  endB <- vector("list", length(set)-1)
  count <- 2
  workNull  <- buildModel(resp,NULL, basis, n)
  endB[[1]] <- calc.MSE_BIC(workNull, dat=dat)
  while(length(set) > 2){ 
    models <- vector("list", length(set))
    k <- 1
    for(z in set)  { # z = "Series1.L1"
      predictors <- set[!set %in% z]
      work <- buildModel(resp, predictors, basis, n)
      models[[k]] <- calc.MSE_BIC(work, dat=dat)
      k <- k + 1
    }
    
    MSEs <- sapply(models, function(zl) zl$MSE)
    ind  <- which.min(MSEs)
    endB[[count]] <- models[[ind]]
    count <- count+1
    set <- models[[ind]]$set
    if(fPath)
      cat(set, "\n") 
  }
  endB
}

############################################################
## INPUT for buildModel():   
##          'predictors' is a set ot covariates
##          'basis' is a basis specification 
## OUTPUT:  list() with builded 'formula' and 'set' for gam fitting.
## Example: predictors = c("lag1","lag3"); n=100
##          basis = "ps"
##          basis = "cr" 
##          buildModel("respY", predictors, basis, n)
##          buildModel(NULL, basis, n)
## Note:
## When evaluating gam(Y ~ s(formula, bs = "ps")) for penalized splines
## the model doesn't work without the function 'p-spline.r'.
############################################################
buildModel <- function(responce, predictors, basis, n){
  if(is.null(predictors)){ 
    res= list(formula=as.formula(paste(responce,"~ 1")),set="1")
    attr(res, "basis") <- basis
    class(res) <- "buildModel"
    return(res)
  }

  if(basis == "ps"){ 
      pred  <- paste("s(",predictors,",bs='",basis,"')",sep="",collapse="+")
      model <- as.formula(paste(responce,"~",pred)) 
    }
  
  if(basis == "cr"){
    k     <- 2                      # tuning constant
    knots <- floor((k*n)^(1/5))     # obtain the number of knots ? problematisch
    pred  <- paste("s(",predictors,",bs='cr',k=",knots,",fx=TRUE)",sep="",collapse="+")
    model <- as.formula(paste(responce,"~",pred)) 
  }    
  res   <- list()
  res$formula <- model
  res$set     <- predictors
  attr(res, "basis") <- basis
  class(res) <- "buildModel"
  return(res) 
}

############################################################################
## INPUT for calc.MSE_BIC(): 
##         'formula_and_set' should be OUTPUT of the function 'buildModel'
############################################################################
calc.MSE_BIC <- function(formula_and_set, dat){
  if(class(formula_and_set)!="buildModel")
    stop("Object must be from class 'buildModel'.")
  require(splines)
  require(mgcv)
  res  <- list()
  fit  <- gam(formula_and_set$formula, data=dat)
  if( attributes(formula_and_set)$basis == "ps" ){
    n    <- length(fit$fitted)
    N    <- sum(fit$edf)
    MSE  <- mean((fit$resid)^2)
    BIC  <- log(MSE) + N/n*log(n)
  }
  if(attributes(formula_and_set)$basis == "cr" ) {
    n    <- length(fit$fitted)
    N    <- 1 + sum(fit$edf) + length(coef(fit)) # sum(fit$edf)==sum(fit$hat)
    MSE  <- mean((fit$resid)^2)
    BIC  <- log(MSE) + N/n*log(n)
  }
  res$fit     <- fit
  res$set     <- formula_and_set$set
   res$MSE     <- MSE
  res$BIC     <- BIC
  return(res)
}
############################################################################
